/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package tsp;

import ilog.concert.IloException;
import ilog.concert.IloNumExpr;
import ilog.concert.IloNumVar;
import ilog.concert.IloNumVarType;
import ilog.cplex.IloCplex;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.Scanner;

/**
 *
 * Inserindo uma formulação geral para o caixeiro viajante. 
 * Os arquivos de entrada são lidos diretamente a partir da pasta ./instances.
 * O arquivo A12.txt foi inicialmente definido. Caso deseje executar para outro arquivo,
 * troque o nome manualmente abaixo em Scanner scan = new Scanner(new File("./instances/A12.txt"))) .
 * No caso da instância A12.txt, ela tem solução ótima e é solucionada em poucos segundos.
 */

public class TSP {

    /**
     * @param args the command line arguments
     * @throws ilog.concert.IloException
     * @throws java.io.FileNotFoundException
     */
    public static void main(String[] args) throws IloException, FileNotFoundException {

    /****  Procedimento básico para leitura de arquivos. 
     * Vc poderá criar outros leitores de arquivo a partir dos passos básicos abaixo
     * 
     */
       
        String otimo;
        int N;
        double[][] C;
        try ( //Fazendo leitura do arquivo
                Scanner scan = new Scanner(new File("./instances/A12.txt"))) {
            scan.nextLine(); // ignore this line
            otimo = scan.nextLine();
            scan.nextLine(); // ignore this line
            N = scan.nextInt();
            //N=3;
            scan.nextLine(); // ignore \n
            scan.nextLine(); // ignore this line
            C = new double[N][N];
            for (int i = 0; i < N; i++) {
                for (int j = 0; j < N; j++) {
                    C[i][j] = scan.nextDouble();
                }
            }
        } // ignore this line

        System.out.println("==========[imprimindo dados]========");
        System.out.println("valor ótimo conhecido = " + otimo);
        System.out.println("cidades = " + N);
        for (int i = 0; i < N; i++) {
            for (int j = 0; j < N; j++) {
                System.out.printf("%10.2g ", C[i][j]);
            }
            System.out.println();
        }
        

        /* Início do código para implementação do modelo geral para o TSP */

        /* classe IloCplex define o objeto (variável) do tipo modelo aqui chamado de cplex 
        O objeto cplex é reponsável por criar variáveis e seus limites, definir restrições,
        parâmetros e resolver o modelo.*/
        IloCplex cplex = new IloCplex();

        //Criação das variáveis binárias  xij como uma matriz do tipo IloNumVar com dimensão N x N
        IloNumVar[][] x = new IloNumVar[N][];
        for (int i = 0; i < N; i++) {
            x[i] = cplex.numVarArray(N, 0, 1, IloNumVarType.Bool);
        }

        //Criação das variáveis contínuas e livres ui como um vetor do tipo IloNumVar com dimensão N
        IloNumVar [] u = cplex.numVarArray(N, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, IloNumVarType.Float);
        

        /*Definindo a função objetivo. 
        Um objeto (variável) do tipo IloNumExpr deve ser criada e iniciada com null para, posteriormente, 
        armazenar a expressão matemática */ 
        IloNumExpr objective = null;
        for (int i = 0; i < N; i++) {
            for (int j = 0; j < N; j++) {        
                if (objective != null){//Deve-se checar para a primeira inserção de expressão matemática
                    objective = cplex.sum(objective, cplex.prod(C[i][j], x[i][j]));
                }else{
                    objective = cplex.prod(C[i][j], x[i][j]);
                }
            }
        }
        //Inserção do objeto objetivo no objeto cplex.
        //Adiciona ao cplex a expressão do objetivo considerando um problema de minimização
        cplex.addMinimize(objective);

        /*Definindo a expressão para a primeira restrição do problema
        Um objeto (variável) do tipo IloNumExpr deve ser criada e iniciada com null para, posteriormente, 
        armazenar a expressão matemática. Como as restrições (1) e (2) são similares, resolvi
        armazenar as restrições em uma matriz do tipo 1xN. Na posição 0 da primeira dimensão (constraints[0][j]), 
        armazeno a j-ésima restrição (1) e, na posição 1 (constraints[1][j]),armazeno a j-ésima restrição (2).
        Lembre-se que temos N restrições de cada tipo*/                     
        IloNumExpr[][] constraints = new IloNumExpr[2][N];//Criando a estrutura matricial
        for (int j = 0; j < N; j++) {//Armazenando as N restrições (1)
            constraints[0][j] = null;
            for (int i = 0; i < N; i++) {        
                if (constraints[0][j] == null){
                    constraints[0][j] = x[i][j];
                }else{
                    constraints[0][j] = cplex.sum(constraints[0][j], x[i][j]);//CLAUDIO: seu codigo tinha apenas essa linha
                }
            }
            /*Adicionamos cada j-ésima restrição ao objeto modelo (cplex).
            Nesse caso, uma restrição de igualdade é inserida*/
            cplex.addEq(constraints[0][j], 1);
        } 
        //Armazenando as N restrições (2)
        for (int i = 0; i < N; i++) {
            constraints[1][i] = null;
            for (int j = 0; j < N; j++) {
                if (constraints[1][i] == null){
                    constraints[1][i] = x[i][j];
                }else{
                    constraints[1][i] = cplex.sum(constraints[1][i], x[i][j]);//CLAUDIO: seu codigo tinha apenas essa linha
                }                         
            }
            /*Adicionamos cada j-ésima restrição ao objeto modelo (cplex).
            Nesse caso, uma restrição de igualdade é inserida*/            
            cplex.addEq(constraints[1][i], 1);
        }

        /*Definindo a expressão para a restrição (3) do TSP. 
        Aqui uma outra forma de armazenar as instruções é utilizada.
        Nesse caso, apenas um objeto do tipo IloNumExpr definido como expr é criado*/
        IloNumExpr expr;
        for (int i = 1; i < N; i++) {            
            for (int j = 1; j < N; j++) {                
                    expr= u[i];
                    expr=cplex.sum(expr,cplex.prod(-1,u[j]));
                    expr=cplex.sum(expr,cplex.prod(N,x[i][j]));                    
                    //Cada uma das restrições é adicionada como uma desigualdade do tipo <= N-1.
                    cplex.addLe(expr, N-1);
            }

        }   
                       
        cplex.exportModel("model.lp");

        if (cplex.solve()) {
            System.out.println("O modelo possui solução");
            System.out.println("Status: " + cplex.getStatus());

            double obj = cplex.getObjValue();
            System.out.println("objetivo =" + obj);

        } else {
            System.out.println("O modelo não possui solução");
            System.out.println("Status: " + cplex.getStatus());
        }

    }

}
